from skimage import filters
import numpy as np
import time
import matplotlib.pyplot as plt

from DensityClust.kc_coord_3d_new import kc_coord_3d, kc_coord_2d
from DensityClust.dist_xyz import dist_xyz
from DensityClust.get_xx import get_xyz


def densityCluster_3d(data, para):
    """
    根据决策图得到聚类中心和聚类中心个数
    :param data: 3D data
    :param para:
        para.rhomin: Minimum density
        para.deltamin: Minimum delta
        para.v_min: Minimum volume
        para.rms: The noise level of the data, used for data truncation calculation
        para.sigma: Standard deviation of Gaussian filtering
    :return:
        NCLUST: number of clusters
        centInd:  centroid index vector
    """
    # 参数初始化
    gradmin = para["gradmin"]
    rhomin = para["rhomin"]
    deltamin = para["deltamin"]
    v_min = para["v_min"]
    rms = para["rms"]
    sigma = para['sigma']
    is_plot = para['is_plot']
    k = 1  # 第1次计算点的邻域大小
    k2 = np.ceil(deltamin).astype(np.int)   # 第2次计算点的邻域大小
    xx = get_xyz(data)  #    xx: 3D data coordinates  坐标原点是 1

    data_filter = filters.gaussian(data, sigma)
    size_x, size_y, size_z = data.shape
    # rho = data_filter.transpose(0, 2, 1).ravel()
    rho = data_filter.flatten()
    rho_Ind = np.argsort(-rho)
    rho_sorted = rho[rho_Ind]
    maxd = size_x + size_y + size_z
    ND = len(rho)

    delta, IndNearNeigh, Gradient = np.zeros(ND, np.float), np.zeros(ND, np.int), np.zeros(ND, np.float)
    delta[rho_Ind[0]] = np.sqrt(size_x ** 2 + size_y ** 2 + size_z ** 2)

    # delta 记录距离，
    # IndNearNeigh
    # 记录：两个密度点的联系 % index of nearest neighbor with higher density
    IndNearNeigh[rho_Ind[0]] = rho_Ind[0]
    t0_ = time.time()
    # 计算 delta, Gradient
    for ii in range(1, ND):
        # 密度降序排序后，即密度第ii大的索引（在rho中）
        # t0 = time.time()
        ordrho_ii = rho_Ind[ii]
        rho_ii = rho_sorted[ii] #第ii大的密度值
        ss = 'litte'
        if rho_ii >= rms:
            delta[ordrho_ii] = maxd
            point_ii_xy = xx[ordrho_ii, :]
            get_value = True  # 判断是否需要在大循环中继续执行，默认需要，一旦在小循环中赋值成功，就不在大循环中运行
            idex, bt = kc_coord_3d(point_ii_xy, size_z, size_y, size_x, k)
            for item_ord, item in zip(idex, bt):
                ordrho_jj = item_ord
                rho_jj = rho[ordrho_jj]  # 根据索引在rho里面取值
                dist_i_j = np.sqrt(((point_ii_xy - item) ** 2).sum())  # 计算两点间的距离
                if dist_i_j <= delta[ordrho_ii]:
                    gradient = (rho_jj - rho_ii) / dist_i_j
                    if gradient >= Gradient[ordrho_ii]:
                        delta[ordrho_ii] = dist_i_j
                        Gradient[ordrho_ii] = gradient
                        IndNearNeigh[ordrho_ii] = ordrho_jj
                        get_value = False

            if get_value:# 表明，在(2 * k + 1) * (2 * k + 1) * (2 * k + 1)的领域中没有找到比该点高，距离最近的点，则在更大的邻域中搜索
                idex, bt = kc_coord_3d(point_ii_xy, size_z, size_y, size_x, k2)
                for item_ord, item in zip(idex, bt):
                    ordrho_jj = item_ord
                    rho_jj = rho[ordrho_jj]  # 根据索引在rho里面取值
                    dist_i_j = np.sqrt(((point_ii_xy - item) ** 2).sum())  # 计算两点间的距离
                    if dist_i_j <= delta[ordrho_ii]:
                        gradient = (rho_jj - rho_ii) / dist_i_j
                        if gradient >= Gradient[ordrho_ii]:
                            delta[ordrho_ii] = dist_i_j
                            Gradient[ordrho_ii] = gradient
                            IndNearNeigh[ordrho_ii] = ordrho_jj
                            get_value = False
            if get_value:
                delta[ordrho_ii] = k2 + 0.0001
                Gradient[ordrho_ii] = -1
                IndNearNeigh[ordrho_ii] = ND
        else:
            IndNearNeigh[ordrho_ii] = ND
        # t1 = time.time()
        # print('%s Elapsed time is %.8f m seconds.\n' % (ss, t1 - t0)*1000)
    delta_sorted = np.sort(-delta) * (-1)
    delta[rho_Ind[0]] = delta_sorted[1]
    t1_ = time.time()
    print('delata, rho and Gradient are calculated, using %.2f seconds' % (t1_-t0_))

    # 根据密度和距离来确定类中心
    NCLUST = 0
    clustInd = -1 * np.ones(ND + 1)
    clust_index = np.intersect1d(np.where(rho > rhomin), np.where(delta > deltamin))

    clust_num = clust_index.shape[0]

    # icl是用来记录第i个类中心在xx中的索引值
    icl = np.zeros(clust_num, dtype=int)
    for ii in range(0, clust_num):
        i = clust_index[ii]
        icl[NCLUST] = i
        NCLUST += 1
        clustInd[i] = NCLUST
    # assignation
    # 将其他非类中心分配到离它最近的类中心中去
    # clustInd = -1
    # 表示该点不是类的中心点，属于其他点，等待被分配到某个类中去
    # 类的中心点的梯度Gradient被指定为 - 1
    # if is_plot == 1:
    #     pass
    for i in range(ND):
        ordrho_i = rho_Ind[i]
        if clustInd[ordrho_i] == -1:    # not centroid
            clustInd[ordrho_i] = clustInd[IndNearNeigh[ordrho_i]]
        else:
            Gradient[ordrho_i] = -1  #将类中心点的梯度设置为-1

    clustVolume = np.zeros(NCLUST)
    for i in range(0, NCLUST):
        clustVolume[i] = clustInd.tolist().count(i + 1)

    # % centInd [类中心点在xx坐标下的索引值，
    # 类中心在centInd的索引值: 代表类别编号]
    centInd = []
    for i, item in enumerate(clustVolume):
        if item >= v_min:
            centInd.append([icl[i], i])
    centInd = np.array(centInd, np.int)

    mask_grad = np.where(Gradient > gradmin)[0]

    # 通过梯度确定边界后，还需要进一步利用最小体积来排除假核
    clustInd_re = -1 * np.ones(ND + 1)   # % 保存最后确定下来的云核的坐标索引
    centInd_re = []  # 保存经过最小体积排出后的类中心信息
    for i, item in enumerate(centInd):
        rho_clust_i = np.zeros(ND)
        # centInd[i, 1] --> item[1] 表示第i个类中心的编号
        index_clust_i = np.where(clustInd == (item[1] + 1))[0]
        index_cc = np.intersect1d(mask_grad, index_clust_i)
        rho_clust_i[index_clust_i] = rho[index_clust_i]
        rho_cc_mean = rho[index_cc].mean() * 0.2
        index_cc_rho = np.where(rho_clust_i > rho_cc_mean)[0]
        index_clust_rho = np.union1d(index_cc, index_cc_rho)
        if len(index_clust_rho) > v_min:
            clustInd_re[index_clust_rho] = i + 1
            centInd_re.append([item[0], i])

    # NCLUST_ = centInd_re.shape[0]
    centInd_re = np.array(centInd_re, np.int)
    return centInd_re, clustInd_re


def get_rho_delta(Gradient, delta, IndNearNeigh, rho_Ind, rho_sorted, rho, rms, maxd, size_x, size_y, size_z, xx, ND):
    # delta 记录距离，
    # IndNearNeigh
    # 记录：两个密度点的联系 % index of nearest neighbor with higher density


    # 计算 delta, Gradient
    for ii in range(1, ND):
        # 密度降序排序后，即密度第ii大的索引（在rho中）
        ordrho_ii = rho_Ind[ii]
        rho_ii = rho_sorted[ii]  # 第ii大的密度值
        if rho_ii >= rms:
            delta[ordrho_ii] = maxd
            point_ii_xy = xx[ordrho_ii, :]
            get_value = True  # 判断是否需要在大循环中继续执行，默认需要，一旦在小循环中赋值成功，就不在大循环中运行
            idex, bt = kc_coord_3d(point_ii_xy, size_z, size_y, size_x, k)
            for item_ord, item in zip(idex, bt):
                ordrho_jj = item_ord

                # ordrho_jj_ = (item[2] - 1) * size_y * size_z + (item[1] - 1) * size_z + item[0]
                # rho_jj_ = data_filter[item[2] - 1, item[1] - 1, item[0] - 1]  #根据坐标在data cube中取值
                rho_jj = rho[ordrho_jj]  # 根据索引在rho里面取值
                # print(rho_jj, rho_jj_)
                # print(ordrho_jj-1, ordrho_jj_)
                dist_i_j = np.sqrt(((point_ii_xy - item) ** 2).sum())  # 计算两点间的距离

                if dist_i_j <= delta[ordrho_ii]:
                    gradient = (rho_jj - rho_ii) / dist_i_j

                    if gradient >= Gradient[ordrho_ii]:
                        delta[ordrho_ii] = dist_i_j
                        Gradient[ordrho_ii] = gradient
                        # IndNearNeigh[ordrho_ii] = (item[2] - 1) * size_y * size_z + (item[1] - 1) * size_z + item[0]
                        IndNearNeigh[ordrho_ii] = ordrho_jj
                        get_value = False
            # for j_ in range(0, bt.shape[0]):
            #     rho_jj = data_filter[bt[j_, 2]-1, bt[j_, 1]-1, bt[j_, 0]-1]
            #     dist_i_j = dist_xyz(delta_ii_xy, bt[j_, :])
            #     gradient = (rho_jj - rho_ii) / dist_i_j
            #     # 通过距离和梯度值的控制可以实现
            #     # 找到“比当前点强度大的点”中最近的点
            #     if dist_i_j <= delta[ordrho_ii] and gradient >= Gradient[ordrho_ii]:
            #         delta[ordrho_ii] = dist_i_j
            #         Gradient[ordrho_ii] = gradient
            #         IndNearNeigh[ordrho_ii] = (bt[j_, 2] - 1) * size_x * size_y + (bt[j_, 1] - 1) * size_x + bt[j_, 0]

            if get_value:
                #  表明，在(2 * k + 1) * (2 * k + 1) * (2 * k + 1)
                # 的领域中没有找到比该点高，距离最近的点，则进行全局搜索
                for jj in range(0, ii, 1):
                    # rho_jj = rho_sorted[jj]
                    ordrho_jj = rho_Ind[jj]
                    point_jj_xy = xx[ordrho_jj, :]

                    dist_i_j = np.sqrt(((point_ii_xy - point_jj_xy) ** 2).sum())
                    gradient = (rho_sorted[jj] - rho_ii) / dist_i_j
                    if dist_i_j <= delta[ordrho_ii]:
                        delta[ordrho_ii] = dist_i_j
                        Gradient[ordrho_ii] = gradient
                        IndNearNeigh[ordrho_ii] = ordrho_jj
        else:
            IndNearNeigh[ordrho_ii] = ND


def densityCluster_3d_multithreading(data, para):
    """
    根据决策图得到聚类中心和聚类中心个数
    :param data: 3D data
    :param para:
        para.rhomin: Minimum density
        para.deltamin: Minimum delta
        para.v_min: Minimum volume
        para.rms: The noise level of the data, used for data truncation calculation
        para.sigma: Standard deviation of Gaussian filtering

    :return:
        NCLUST: number of clusters
        centInd:  centroid index vector
    """
    # 参数初始化
    gradmin = para["gradmin"]
    rhomin = para["rhomin"]
    deltamin = para["deltamin"]
    v_min = para["v_min"]
    rms = para["rms"]
    sigma = para['sigma']
    is_plot = para['is_plot']
    k = 2 # 计算点的邻域大小
    xx = get_xyz(data)  # xx: 3D data coordinates  坐标原点是 1

    data_filter = filters.gaussian(data, sigma)
    size_x, size_y, size_z = data.shape
    # rho = data_filter.transpose(0, 2, 1).ravel()
    rho = data_filter.flatten()
    rho_Ind = np.argsort(-rho)
    rho_sorted = rho[rho_Ind]
    maxd = size_x + size_y + size_z
    ND = len(rho)

    delta, IndNearNeigh, Gradient = np.zeros(ND, np.float), np.zeros(ND, np.int), np.zeros(ND, np.float)
    delta[rho_Ind[0]] = np.sqrt(size_x ** 2 + size_y ** 2 + size_z ** 2)
    IndNearNeigh[rho_Ind[0]] = rho_Ind[0]
    t0 = time.time()
    get_rho_delta(Gradient, delta, IndNearNeigh, rho_Ind, rho_sorted, rho, rms, maxd, size_x, size_y, size_z, xx, ND)
    delta_sorted = np.sort(-delta) * -1
    delta[rho_Ind[0]] = delta_sorted[1]
    t1 = time.time()
    print('delata, rho and Gradient are calculated, using %.2f seconds' % (t1-t0))

    # 根据密度和距离来确定类中心
    NCLUST = 0
    clustInd = -1 * np.ones(ND + 1)
    clust_index = np.intersect1d(np.where(rho > rhomin), np.where(delta > deltamin))

    clust_num = clust_index.shape[0]

    # icl是用来记录第i个类中心在xx中的索引值
    icl = np.zeros(clust_num, dtype=int)
    for ii in range(0, clust_num):
        i = clust_index[ii]
        icl[NCLUST] = i
        NCLUST += 1
        clustInd[i] = NCLUST

    # assignation
    # 将其他非类中心分配到离它最近的类中心中去
    # clustInd = -1
    # 表示该点不是类的中心点，属于其他点，等待被分配到某个类中去
    # 类的中心点的梯度Gradient被指定为 - 1
    if is_plot == 1:
        delta = delta / max(delta[:])
        # figure
    for i in range(0, ND):
        ordrho_i = rho_Ind[i]
        if clustInd[ordrho_i] == -1:# not centroid
            clustInd[ordrho_i] = clustInd[IndNearNeigh[ordrho_i]]
        else:
            Gradient[ordrho_i] = -1  #将类中心点的梯度设置为-1

    clustVolume = np.zeros(NCLUST)
    for i in range(0, NCLUST):
        clustVolume[i] = clustInd.tolist().count(i + 1)

    # % centInd [类中心点在xx坐标下的索引值，
    # 类中心在centInd的索引值: 代表类别编号]
    centInd = []
    for i, item in enumerate(clustVolume):
        if item >= v_min:
            centInd.append([icl[i], i])
    centInd = np.array(centInd, np.int)

    mask_grad = np.where(Gradient > gradmin)[0]

    # 通过梯度确定边界后，还需要进一步利用最小体积来排除假核
    clustInd_re = -1 * np.ones(ND + 1)   # % 保存最后确定下来的云核的坐标索引
    centInd_re = []  # 保存经过最小体积排出后的类中心信息
    for i, item in enumerate(centInd):
        rho_clust_i = np.zeros(ND)
        # centInd[i, 1] --> item[1] 表示第i个类中心的编号
        index_clust_i = np.where(clustInd == (item[1] + 1))[0]
        index_cc = np.intersect1d(mask_grad, index_clust_i)
        rho_clust_i[index_clust_i] = rho[index_clust_i]
        rho_cc_mean = rho[index_cc].mean() * 0.2
        index_cc_rho = np.where(rho_clust_i > rho_cc_mean)[0]
        index_clust_rho = np.union1d(index_cc, index_cc_rho)
        if len(index_clust_rho) > v_min:
            clustInd_re[index_clust_rho] = i + 1
            centInd_re.append(item)

    # NCLUST_ = centInd_re.shape[0]
    centInd_re = np.array(centInd_re, np.int)
    return centInd_re, clustInd_re


def densityCluster_2d(data, para):
    """
    根据决策图得到聚类中心和聚类中心个数
    :param data: 2D data
    :param para:
        para.rhomin: Minimum density
        para.deltamin: Minimum delta
        para.v_min: Minimum volume
        para.rms: The noise level of the data, used for data truncation calculation
        para.sigma: Standard deviation of Gaussian filtering

    :return:
        NCLUST: number of clusters
        centInd:  centroid index vector
    """
    # 参数初始化
    gradmin = para["gradmin"]
    rhomin = para["rhomin"]
    deltamin = para["deltamin"]
    v_min = para["v_min"]
    rms = para["rms"]
    sigma = para['sigma']
    is_plot = para['is_plot']
    k = 1   # 计算点的邻域大小
    xx = get_xyz(data)  # xx: 2D data coordinates  坐标原点是 1

    data_filter = filters.gaussian(data, sigma)
    size_x, size_y = data.shape
    rho = data_filter.flatten()
    rho_Ind = np.argsort(-rho)
    rho_sorted = rho[rho_Ind]
    maxd = size_x + size_y
    ND = len(rho)

    # delta 记录距离， # IndNearNeigh 记录：两个密度点的联系 % index of nearest neighbor with higher density
    delta, IndNearNeigh, Gradient = np.zeros(ND, np.float), np.zeros(ND, np.int), np.zeros(ND, np.float)

    delta[rho_Ind[0]] = np.sqrt(size_x ** 2 + size_y ** 2)
    IndNearNeigh[rho_Ind[0]] = rho_Ind[0]

    t0 = time.time()
    # 计算 delta, Gradient
    for ii in range(1, ND):
        # 密度降序排序后，即密度第ii大的索引（在rho中）
        ordrho_ii = rho_Ind[ii]
        rho_ii = rho_sorted[ii]   # 第ii大的密度值
        if rho_ii >= rms:
            delta[ordrho_ii] = maxd
            delta_ii_xy = xx[ordrho_ii, :]

            bt = kc_coord_2d(delta_ii_xy, size_y, size_x, k)
            for item in bt:
                rho_jj = data_filter[item[1] - 1, item[0] - 1]
                dist_i_j = np.sqrt(((delta_ii_xy - item) ** 2).sum())  # 计算两点间的距离
                gradient = (rho_jj - rho_ii) / dist_i_j

                if dist_i_j <= delta[ordrho_ii] and gradient >= Gradient[ordrho_ii]:
                    delta[ordrho_ii] = dist_i_j
                    Gradient[ordrho_ii] = gradient
                    IndNearNeigh[ordrho_ii] = (item[1] - 1) * size_y + item[0] - 1

            if delta[ordrho_ii] == maxd:
                # 表明，在(2 * k + 1) * (2 * k + 1) * (2 * k + 1)
                # 的领域中没有找到比该点高，距离最近的点，则进行全局搜索
                for jj in range(0, ii, 1):
                    rho_jj = rho_sorted[jj]
                    ordrho_jj = rho_Ind[jj]
                    delta_jj_xy = xx[ordrho_jj, :]
                    dist_i_j = np.sqrt(((delta_ii_xy - delta_jj_xy) ** 2).sum())
                    gradient = (rho_jj - rho_ii) / dist_i_j
                    if dist_i_j <= delta[ordrho_ii]:
                        delta[ordrho_ii] = dist_i_j
                        Gradient[ordrho_ii] = gradient
                        IndNearNeigh[ordrho_ii] = ordrho_jj
        else:
            IndNearNeigh[ordrho_ii] = ND

    delta_sorted = np.sort(-delta) * (-1)
    delta[rho_Ind[0]] = delta_sorted[1]
    t1 = time.time()
    print('delata, rho and Gradient are calculated, using %.2f seconds' % (t1-t0))

    # 根据密度和距离来确定类中心
    NCLUST = 0
    clustInd = -1 * np.ones(ND + 1)
    clust_index = np.intersect1d(np.where(rho > rhomin), np.where(delta > deltamin))

    clust_num = clust_index.shape[0]
    print(clust_num)

    # icl是用来记录第i个类中心在xx中的索引值
    icl = np.zeros(clust_num, dtype=int)
    for ii in range(0, clust_num):
        i = clust_index[ii]
        icl[NCLUST] = i
        NCLUST += 1
        clustInd[i] = NCLUST

    # assignation
    # 将其他非类中心分配到离它最近的类中心中去
    # clustInd = -1
    # 表示该点不是类的中心点，属于其他点，等待被分配到某个类中去
    # 类的中心点的梯度Gradient被指定为 - 1
    if is_plot == 1:

        plt.scatter(rho, delta, marker='.')
        plt.show()

    for i in range(ND):
        ordrho_i = rho_Ind[i]
        if clustInd[ordrho_i] == -1:    # not centroid
            clustInd[ordrho_i] = clustInd[IndNearNeigh[ordrho_i]]
        else:
            Gradient[ordrho_i] = -1  # 将类中心点的梯度设置为-1

    clustVolume = np.zeros(NCLUST)
    for i in range(NCLUST):
        clustVolume[i] = clustInd.tolist().count(i + 1)

    # % centInd [类中心点在xx坐标下的索引值，
    # 类中心在centInd的索引值: 代表类别编号]
    centInd = []
    for i, item in enumerate(clustVolume):
        if item >= v_min:
            centInd.append([icl[i], i])
    centInd = np.array(centInd, np.int)
    print(centInd.shape[0])
    mask_grad = np.where(Gradient > gradmin)[0]

    # 通过梯度确定边界后，还需要进一步利用最小体积来排除假核
    clustInd_re = -1 * np.ones(ND + 1)   # % 保存最后确定下来的云核的坐标索引
    centInd_re = []  # 保存经过最小体积排出后的类中心信息
    for i, item in enumerate(centInd):
        rho_clust_i = np.zeros(ND)
        # centInd[i, 1] --> item[1] 表示第i个类中心的编号
        index_clust_i = np.where(clustInd == (item[1] + 1))[0]
        index_cc = np.intersect1d(mask_grad, index_clust_i)
        rho_clust_i[index_clust_i] = rho[index_clust_i]
        if len(index_cc) > 0:
            rho_cc_mean = rho[index_cc].mean() * 0.2
        else:
            rho_cc_mean = rms
        index_cc_rho = np.where(rho_clust_i > rho_cc_mean)[0]
        index_clust_rho = np.union1d(index_cc, index_cc_rho)

        if len(index_clust_rho) > v_min:
            # print(i + 1)
            clustInd_re[index_clust_rho] = i + 1
            centInd_re.append([item[0], i])
            # centInd_re.append(item)

    # NCLUST_ = centInd_re.shape[0]
    centInd_re = np.array(centInd_re, np.int)
    return centInd_re, clustInd_re


def densityCluster_2d_new(data, para):
    """
    根据决策图得到聚类中心和聚类中心个数
    :param data: 2D data
    :param para:
        para.rhomin: Minimum density
        para.deltamin: Minimum delta
        para.v_min: Minimum volume
        para.rms: The noise level of the data, used for data truncation calculation
        para.sigma: Standard deviation of Gaussian filtering

    :return:
        NCLUST: number of clusters
        centInd:  centroid index vector
    """
    # 参数初始化
    gradmin = para["gradmin"]
    rhomin = para["rhomin"]
    deltamin = para["deltamin"]
    v_min = para["v_min"]
    rms = para["rms"]
    sigma = para['sigma']
    is_plot = para['is_plot']
    k = 2 # 计算点的邻域大小
    xx = get_xyz(data)  # xx: 2D data coordinates  坐标原点是 1

    data_filter = filters.gaussian(data, sigma)
    size_x, size_y = data.shape
    rho = data_filter.flatten()
    rho_Ind = np.argsort(-rho)
    rho_sorted = rho[rho_Ind]
    maxd = size_x + size_y
    ND = len(rho)

    delta, IndNearNeigh, Gradient = np.zeros(ND, np.float), np.zeros(ND, np.int), np.zeros(ND, np.float)
    delta[rho_Ind[0]] = np.sqrt(size_x ** 2 + size_y ** 2)

    # delta 记录距离，
    # IndNearNeigh
    # 记录：两个密度点的联系 % index of nearest neighbor with higher density
    IndNearNeigh[rho_Ind[0]] = rho_Ind[0]
    t0 = time.time()
    # 计算 delta, Gradient
    for ii in range(1, ND):
        # 密度降序排序后，即密度第ii大的索引（在rho中）
        ordrho_ii = rho_Ind[ii]
        rho_ii = rho_sorted[ii] #第ii大的密度值
        if rho_ii >= rms:
            delta[ordrho_ii] = maxd
            delta_ii_xy = xx[ordrho_ii, :]

            bt = kc_coord_2d(delta_ii_xy, size_y, size_x, k)
            for item in bt:
                rho_jj = data_filter[item[1] - 1, item[0] - 1]
                dist_i_j = dist_xyz(delta_ii_xy, item)
                gradient = (rho_jj - rho_ii) / dist_i_j

                if dist_i_j <= delta[ordrho_ii] and gradient >= Gradient[ordrho_ii]:
                    delta[ordrho_ii] = dist_i_j
                    Gradient[ordrho_ii] = gradient
                    IndNearNeigh[ordrho_ii] = (item[1] - 1) * size_y + item[0] - 1

            if delta[ordrho_ii] == maxd:
                # 表明，在(2 * k + 1) * (2 * k + 1) * (2 * k + 1)
                # 的领域中没有找到比该点高，距离最近的点，则进行全局搜索
                for jj in range(0, ii, 1):
                    rho_jj = rho_sorted[jj]
                    ordrho_jj = rho_Ind[jj]
                    delta_jj_xy = xx[ordrho_jj, :]
                    temp = delta_ii_xy - delta_jj_xy
                    dist_i_j = np.sqrt((temp ** 2).sum())
                    gradient = (rho_jj - rho_ii) / dist_i_j
                    if dist_i_j <= delta[ordrho_ii]:
                        delta[ordrho_ii] = dist_i_j
                        Gradient[ordrho_ii] = gradient
                        IndNearNeigh[ordrho_ii] = ordrho_jj
        else:
            IndNearNeigh[ordrho_ii] = ND

    delta_sorted = np.sort(-delta) * (-1)
    # delta[rho_Ind[0]] = delta_sorted[1]
    t1 = time.time()
    print('delata, rho and Gradient are calculated, using %.2f seconds' % (t1-t0))

    # 根据密度和距离来确定类中心
    NCLUST = 0
    clustInd = -1 * np.ones(ND + 1)
    clust_index = np.intersect1d(np.where(rho > rhomin), np.where(delta > deltamin))

    clust_num = clust_index.shape[0]
    print(clust_num)

    # icl是用来记录第i个类中心在xx中的索引值
    icl = np.zeros(clust_num, dtype=int)
    for ii in range(0, clust_num):
        i = clust_index[ii]
        icl[NCLUST] = i
        NCLUST += 1
        clustInd[i] = NCLUST

    # assignation
    # 将其他非类中心分配到离它最近的类中心中去
    # clustInd = -1
    # 表示该点不是类的中心点，属于其他点，等待被分配到某个类中去
    # 类的中心点的梯度Gradient被指定为 - 1
    if is_plot == 1:
        delta = delta / max(delta[:])
        # figure
    for i in range(0, ND):
        ordrho_i = rho_Ind[i]
        if clustInd[ordrho_i] == -1:# not centroid
            clustInd[ordrho_i] = clustInd[IndNearNeigh[ordrho_i]]
        else:
            Gradient[ordrho_i] = -1  #将类中心点的梯度设置为-1

    clustVolume = np.zeros(NCLUST)
    for i in range(0, NCLUST):
        clustVolume[i] = clustInd.tolist().count(i + 1)

    # % centInd [类中心点在xx坐标下的索引值，
    # 类中心在centInd的索引值: 代表类别编号]
    centInd = []
    for i, item in enumerate(clustVolume):
        if item >= v_min:
            centInd.append([icl[i], i])
    centInd = np.array(centInd, np.int)

    mask_grad = np.where(Gradient > gradmin)[0]

    # 通过梯度确定边界后，还需要进一步利用最小体积来排除假核
    clustInd_re = -1 * np.ones(ND + 1)   # % 保存最后确定下来的云核的坐标索引
    centInd_re = []  # 保存经过最小体积排出后的类中心信息
    for i, item in enumerate(centInd):
        rho_clust_i = np.zeros(ND)
        # centInd[i, 1] --> item[1] 表示第i个类中心的编号
        index_clust_i = np.where(clustInd == (item[1] + 1))[0]
        index_cc = np.intersect1d(mask_grad, index_clust_i)
        rho_clust_i[index_clust_i] = rho[index_clust_i]
        rho_cc_mean = rho[index_cc].mean() * 0.2
        index_cc_rho = np.where(rho_clust_i > rho_cc_mean)[0]
        index_clust_rho = np.union1d(index_cc, index_cc_rho)
        if len(index_clust_rho) > v_min:
            clustInd_re[index_clust_rho] = i + 1

            centInd_re.append(item)

    # NCLUST_ = centInd_re.shape[0]
    centInd_re = np.array(centInd_re, np.int)
    return centInd_re, clustInd_re


if __name__ == '__main__':
    pass


import numpy as np
from skimage import measure, morphology
from scipy import ndimage
import matplotlib.pyplot as plt
from DensityClust import get_xx, make_plot, tools



def make_tri_plot(data):
    fig, (ax0, ax1, ax2) = plt.subplots(1, 3, figsize=(8, 6))
    ax0.imshow(data.sum(axis=0), cmap='gray',)
    ax1.imshow(data.sum(axis=1), cmap='gray')
    ax2.imshow(data.sum(axis=2), cmap='gray')

    fig.tight_layout()
    plt.xticks([])
    plt.yticks([])
    plt.show()


def extroclump_parameters(clustInd, centInd, data):
    """
    计算根据LDC得到的聚类的云核参数

    :param clustInd: 聚类的编号，同一个类用一个相同的数字标记
    :param centInd: centInd: centInd = [centIndex, clust_id] 代表聚类中心点在data中的索引以及聚类的编号(ID)
    :param data: 3D data
    :return:
    output:
        Peak1,Peak2,Peak3,Cen1,Cen2,Cen3,Size1,Size2,Size3,Peak,Sum,Volume
        Size1, Size2, Size3  为半高全宽(FWHM)
    out: clumps data
    mask: mask
    """
    
    NClust = centInd.shape[0]
    xx = get_xx.get_xyz(data)  # xx: 3D data coordinates  xx坐标系的坐标原点为1
    dim = data.ndim
    mask = np.zeros_like(data, dtype=np.int)
    out = np.zeros_like(data, dtype=np.float)

    clustSum = np.zeros([NClust, 1], dtype=np.float)
    clustVolume = np.zeros([NClust, 1], dtype=np.int)
    clustPeak = np.zeros([NClust, 1], dtype=np.float)
    clump_Cen = np.zeros([NClust, dim], dtype=np.float)
    clustSize = np.zeros([NClust, dim], dtype=np.float)
    if dim == 3:
        clump_Peak = xx[centInd[:, 0], :]   # xx坐标系的坐标原点为1

        for i in range(NClust):

            cl_1_index_ = xx[np.where(clustInd == (centInd[i, 1]) + 1)[0], :] - 1  # -1 是为了在data里面用索引取值(从0开始)
            # clustInd  标记的点的编号是从1开始，  没有标记的点的编号为-1
            clustNum = cl_1_index_.shape[0]

            cl_i = np.zeros_like(data, np.int)
            for j, item in enumerate(cl_1_index_):
                cl_i[item[2], item[1], item[0]] = 1
            # 形态学处理
            L = measure.label(cl_i)   # Labeled input image. Labels with value 0 are ignored.
            STATS = measure.regionprops(L)

            Ar_sum = []
            for region in STATS:
                coords = region.coords  # 经过验证，坐标原点为0
                temp = 0
                for j, item in enumerate(coords):
                    temp += data[item[0], item[1], item[2]]
                Ar_sum.append(temp)
            Ar = np.array(Ar_sum)
            ind = np.where(Ar == Ar.max())[0]
            L[L != ind + 1] = 0
            cl_i = L / (ind[0] + 1)
            coords_maxAr = STATS[ind[0]].coords  # 最大的连通域对应的坐标

            clump_i_ = np.zeros(coords_maxAr.shape[0])
            for j, item in enumerate(coords_maxAr):
                clump_i_[j] = data[item[0], item[1], item[2]]

            clump_i = data * cl_i
            out = out + clump_i
            mask = mask + cl_i * (i + 1)
            clustsum = clump_i_.sum() + 0.0001  # 加一个0.0001 防止分母为0
            coords_maxAr = coords_maxAr[:, [2, 1, 0]]  # 和xx坐标系保持一致
            clump_Cen[i, :] = np.matmul(clump_i_, coords_maxAr) / clustsum  # 这里的坐标，坐标原点为0
            clustVolume[i, 0] = clustNum
            clustSum[i, 0] = clustsum
            clustPeak[i, 0] = data[clump_Peak[i, 2] - 1, clump_Peak[i, 1] - 1, clump_Peak[i, 0] - 1]

            x_i = coords_maxAr - clump_Cen[i, :]
            clustSize[i, :] = 2.3548 * np.sqrt((np.matmul(clump_i_, x_i ** 2) / clustsum)
                                      - (np.matmul(clump_i_, x_i) / clustsum) ** 2)

    else:
        clustSum, clustVolume, clustPeak = np.zeros([NClust, 1]), np.zeros([NClust, 1]), np.zeros([NClust, 1])
        clump_Cen, clustSize = np.zeros([NClust, dim]), np.zeros([NClust, dim])
        clump_Peak = xx[centInd[:, 0], :]

        for i in range(0, NClust):

            cl_i = np.zeros_like(data, np.int)
            cl_1_index_ = xx[np.where(clustInd == (centInd[i, 1]) + 1)[0], :] - 1  # -1 是为了在data里面用索引取值(从0开始)
            clustNum = cl_1_index_.shape[0]

            for j, item in enumerate(cl_1_index_):
                # print(item)
                cl_i[item[1], item[0]] = 1  # (i + 1) 是云核的编号  从1开始

            print(i)
            L = measure.label(cl_i)  # 给连通域打标签
            STATS = measure.regionprops(L)  # 统计连通域的信息
            Ar_sum = []
            for region in STATS:
                # Ar.append(region.area)
                coords = region.coords
                coords = coords[:, [1, 0]]
                temp = 0
                for j, item in enumerate(coords):
                    temp += data[item[1], item[0]]
                Ar_sum.append(temp)
            Ar = np.array(Ar_sum)
            ind = np.where(Ar == Ar.max())[0]
            L[L != ind + 1] = 0
            cl_i = L / (ind[0] + 1)
            coords = STATS[ind[0]].coords  # 最大的连通域对应的坐标

            coords = coords[:, [1, 0]]
            clump_i_ = np.zeros(coords.shape[0])
            for j, item in enumerate(coords):
                clump_i_[j] = data[item[1], item[0]]

            clustsum = sum(clump_i_) + 0.0001  # 加一个0.0001 防止分母为0
            clump_Cen[i, :] = np.matmul(clump_i_, coords) / clustsum
            clustVolume[i, 0] = clustNum
            clustSum[i, 0] = clustsum
            clustPeak[i, 0] = data[clump_Peak[i, 1] - 1, clump_Peak[i, 0] - 1]
            x_i = coords - clump_Cen[i, :]
            clustSize[i, :] = 2.3548 * np.sqrt((np.matmul(clump_i_, x_i ** 2) / clustsum)
                                      - (np.matmul(clump_i_, x_i) / clustsum) ** 2)
            clump_i = data * cl_i
            out = out + clump_i
            mask = mask + cl_i * (i + 1)

    clump_Cen = clump_Cen + 1  # python坐标原点是从0开始的，在这里整体加1，改为以1为坐标原点
    id_clumps = np.array([item + 1 for item in range(NClust)], np.int).T
    id_clumps = id_clumps.reshape([NClust, 1])

    outcat = np.column_stack((id_clumps, clump_Peak, clump_Cen, clustSize, clustPeak, clustSum, clustVolume))

    return outcat, mask, out


# def extroclump_parameters(centNum, xxyy, co_real, centInd, origin_data):
#     dim = len(origin_data.shape)
#     NCLUST = centNum.shape[0]
#     [xres, yres] = origin_data.shape
#     clustSum = np.zeros(NCLUST)
#     clustVolume = np.zeros(NCLUST)
#     clustPeak = np.zeros(NCLUST)
#     clumpCenter = np.zeros([NCLUST, dim])
#     clustSize = np.zeros([NCLUST, dim])
#
#     clumpPeakIndex = xxyy[centInd]
#     describe = {}
#     for i in range(NCLUST):
#         clust_coords = xxyy[co_real == centNum[i]]
#         clustSum[i] = origin_data[clust_coords[:, 0], clust_coords[:, 1]].sum()
#         clustVolume[i] = clust_coords.shape[0]
#         clustPeak[i] = origin_data[clumpPeakIndex[i][0], clumpPeakIndex[i][1]]
#         x_center = np.sum((clust_coords[:, 0] + 1) * origin_data[clust_coords[:, 0], clust_coords[:, 1]]) / clustSum[
#             i] - 1
#         y_center = np.sum((clust_coords[:, 1] + 1) * origin_data[clust_coords[:, 0], clust_coords[:, 1]]) / clustSum[
#             i] - 1
#         clumpCenter[i] = [x_center, y_center]
#         center_delta = clust_coords - clumpCenter[i]
#         od = origin_data[clust_coords[:, 0], clust_coords[:, 1]]
#         clustSize[i] = (np.array(np.mat(od) * np.array(np.mat(center_delta)) ** 2 / clustSum[i]) \
#                         - np.array(np.mat(od) * np.mat(center_delta) / clustSum[i]) ** 2) ** (1 / 2)
#     describe['clustSum'] = clustSum
#     describe['clustVolume'] = clustVolume
#     describe['clumpPeakIndex'] = clumpPeakIndex
#     describe['clustPeak'] = clustPeak
#     describe['clumpCenter'] = clumpCenter
#     describe['clustSize'] = clustSize
#     return describe

if __name__ == '__main__':
    import astropy.io.fits as fits
    centInd = fits.getdata('centInd.fits')
    data = fits.getdata('data.fits')
    clustInd = fits.getdata('clustInd.fits')

    outcat, mask, out = extroclump_parameters(clustInd, centInd, data)
    outcat_name = 'data.fits'.replace('.fits', '_LDC1.txt')
    tools.save_outcat(outcat_name=outcat_name, outcat=outcat)
    make_plot.make_plot(outcat_name, out)
    mask1 = mask.copy()
    mask1[mask1!=67] = 0
    make_plot.make_plot(outcat_name, out * mask1/67)



import numpy as np


def distx(kc1, kc2, xx):
    if xx.shape[1] == 3:
        distance = np.sqrt((xx[kc1, 0] - xx[kc2, 0]) ** 2 + (xx[kc1, 1] - xx[kc2, 1]) ** 2 +
                           (xx[kc1, 2] - xx[kc2, 2]) ** 2)
    else:
        distance = np.sqrt((xx[kc1, 0] - xx[kc2, 0]) ** 2 + (xx[kc1, 1] - xx[kc2, 1]) ** 2)
    return distance



import os
import numpy as np
import astropy.io.fits as fits
import time
import pandas as pd
from astropy import wcs
from DensityClust import make_plot, tools
from DensityClust.densityCluster_3d import densityCluster_3d, densityCluster_2d
from DensityClust.extroclump_parameters import extroclump_parameters


def get_wcs(data_name):
    """
    得到wcs信息
    :param data_name: fits文件
    :return:
    data_wcs
    """
    data_header = fits.getheader(data_name)
    keys = data_header.keys()
    key = [k for k in keys if k.endswith('4')]
    [data_header.remove(k) for k in key]
    data_header.remove('VELREF')
    data_wcs = wcs.WCS(data_header)
    return data_wcs


def change_pix2word(data_wcs, outcat):
    """
    将算法检测的结果(像素单位)转换到天空坐标系上去
    :param data_wcs: 头文件得到的wcs
    :param outcat: 算法检测核表
    :return:
    outcat_wcs
    ['ID', 'Peak1', 'Peak2', 'Peak3', 'Cen1', 'Cen2', 'Cen3', 'Size1', 'Size2', 'Size3', 'Peak', 'Sum', 'Volume']
    """
    peak1, peak2, peak3 = data_wcs.all_pix2world(outcat['Peak1'], outcat['Peak2'], outcat['Peak3'], 1)
    clump_Peak = np.column_stack([peak1, peak2, peak3 / 1000])
    cen1, cen2, cen3 = data_wcs.all_pix2world(outcat['Cen1'], outcat['Cen2'], outcat['Cen3'], 1)
    clump_Cen = np.column_stack([cen1, cen2, cen3 / 1000])
    size1, size2, size3 = np.array([outcat['Size1'] * 30, outcat['Size2'] * 30, outcat['Size3'] * 166.6])
    clustSize = np.column_stack([size1, size2, size3])
    clustPeak, clustSum, clustVolume = np.array([outcat['Peak'], outcat['Sum'], outcat['Volume']])

    id_clumps = []  # G017.558+00.150+020.17  分别表示：银经：17.558°， 银纬：0.15°，速度：20.17km/s
    for item_l, item_b, item_v in zip(cen1, cen2, cen3 / 1000):
        str_l = 'G' + ('%.03f' % item_l).rjust(7, '0')
        if item_b < 0:
            str_b = '-' + ('%.03f' % abs(item_b)).rjust(6, '0')
        else:
            str_b = '+' + ('%.03f' % abs(item_b)).rjust(6, '0')
        if item_v < 0:
            str_v = '-' + ('%.03f' % abs(item_v)).rjust(6, '0')
        else:
            str_v = '+' + ('%.03f' % abs(item_v)).rjust(6, '0')
        id_clumps.append(str_l + str_b + str_v)
    id_clumps = np.array(id_clumps)

    outcat_wcs = np.column_stack((id_clumps, clump_Peak, clump_Cen, clustSize, clustPeak, clustSum, clustVolume))
    return outcat_wcs


def get_outcat_local(outcat):
    """
    返回局部区域的检测结果：
    原始图为120*120  局部区域为30-->90, 30-->90 左开右闭
    :param outcat:
    :return:
    """
    # outcat = pd.read_csv(txt_name, sep='\t')
    cen1_min = 30
    cen1_max = 90
    cen2_min = 30
    cen2_max = 90
    aa = outcat.loc[outcat['Cen1'] > cen1_min]
    aa = aa.loc[outcat['Cen1'] <= cen1_max]
    aa = aa.loc[outcat['Cen2'] > cen2_min]
    aa = aa.loc[outcat['Cen2'] <= cen2_max]
    return aa


def localDenCluster(data_name, para=None, mask_name=None, outcat_name=None, outcat_wcs_name=None):
    data_port_dir = data_name.replace('.fits', '')
    if para is None:
        para = {"gradmin": 0.01, "rhomin": 0.8, "deltamin": 4, "v_min": 27, "rms": 0.46, "sigma": 0.6, "is_plot": 0}

    if not os.path.exists(data_port_dir):
        os.mkdir(data_port_dir)

    data = fits.getdata(data_name)
    data[np.isnan(data)] = 0
    data_ndim = data.ndim
    if data_ndim == 2:
        print("2d")
        centInd, clustInd = densityCluster_2d(data, para)
    elif data_ndim == 3:
        print("3d data")
        t0 = time.time()
        centInd, clustInd = densityCluster_3d(data, para)
        t1 = time.time()
        print('LDC time: %0.2f' % (t1-t0))
    else:
        centInd, clustInd = None, None

    if centInd is not None:
        outcat, mask, out = extroclump_parameters(clustInd, centInd, data)
        if outcat_name is None:
            outcat_name = os.path.join(data_port_dir, 'LDC_outcat.txt')
        if mask_name is None:
            mask_name = os.path.join(data_port_dir, 'LDC_mask.fits')
        if outcat_wcs_name is None:
            outcat_wcs_name = os.path.join(data_port_dir, 'LDC_outcat_wcs.txt')

        if os.path.isfile(mask_name):
            os.remove(mask_name)
            fits.writeto(mask_name, mask)
        else:
            fits.writeto(mask_name, mask)

        tools.save_outcat(outcat_name=outcat_name, outcat=outcat)

        outcat = pd.read_csv(outcat_name, sep='\t')
        data_wcs = get_wcs(data_name)
        outcat_wcs = change_pix2word(data_wcs, outcat)

        tools.save_outcat(outcat_name=outcat_wcs_name, outcat=outcat_wcs)





import pandas as pd

# https://stackoverflow.com/questions/16490261/python-pandas-write-dataframe-to-fixed-width-file-to-fwf

from tabulate import tabulate
import numpy as np


def to_fwf(df, fname):
    content = tabulate(df.values.tolist(), list(df.columns), tablefmt="plain")
    open(fname, "w").write(content)


def save_outcat(outcat_name, outcat):
    outcat_colums = outcat.shape[1]
    pd.DataFrame.to_fwf = to_fwf
    if outcat_colums == 10:
        detected_outcat_name = outcat_name
        table_title = ['ID', 'Peak1', 'Peak2', 'Cen1', 'Cen2', 'Size1', 'Size2', 'Peak', 'Sum',
                       'Volume']
        dataframe = pd.DataFrame(outcat, columns=table_title)
        dataframe = dataframe.round({'ID': 0, 'Peak1': 0, 'Peak2': 0, 'Cen1': 3, 'Cen2': 3,
                                     'Size1': 3, 'Size2': 3, 'Peak': 3, 'Sum': 3, 'Volume': 3})
        dataframe.to_csv(detected_outcat_name, sep='\t', index=False)
        # dataframe.to_fwf(detected_outcat_name)


    if outcat_colums == 13:
        print(13)
        detected_outcat_name = outcat_name
        table_title = ['ID', 'Peak1', 'Peak2', 'Peak3', 'Cen1', 'Cen2', 'Cen3', 'Size1', 'Size2', 'Size3', 'Peak', 'Sum',
                       'Volume']
        dataframe = pd.DataFrame(outcat, columns=table_title)
        dataframe = dataframe.round({'ID': 0, 'Peak1': 0, 'Peak2': 0, 'Peak3': 0, 'Cen1': 3, 'Cen2': 3, 'Cen3': 3,
                                     'Size1': 3, 'Size2': 3, 'Size3': 3, 'Peak': 3, 'Sum': 3, 'Volume': 3})
        dataframe.to_csv(detected_outcat_name, sep='\t', index=False)
        # dataframe.to_fwf(detected_outcat_name)

    if outcat_colums == 11:
        fit_outcat_name = outcat_name
        fit_outcat = outcat
        table_title = ['ID', 'Peak1', 'Peak2', 'Cen1', 'Cen2', 'Size1', 'Size2', 'theta', 'Peak',
                       'Sum', 'Volume']
        dataframe = pd.DataFrame(fit_outcat, columns=table_title)
        dataframe = dataframe.round(
            {'ID': 0, 'Peak1': 3, 'Peak2': 3, 'Cen1': 3, 'Cen2': 3, 'Size1': 3, 'Size2': 3, 'theta': 3, 'Peak': 3,
             'Sum': 3, 'Volume': 3})
        dataframe.to_csv(fit_outcat_name, sep='\t', index=False)

        # dataframe.to_fwf(fit_outcat_name)


if __name__ == '__main__':
    pd.DataFrame.to_fwf = to_fwf

    # %%
    # df = pd.DataFrame({"ID": [1, 2, 3, 4, 50000], "a": np.random.rand(5), "b": [1.0, 2.01, 3.02, 4.03, 25.006]})
    df = pd.read_csv(r'F:\LDC_python\detection\R2_data\data_9\0175+000\0175+000_L\LDC_outcat_wcs.txt', sep='\t')
    # pd.DataFrame.to_fwf = to_fwf

    # %%
    # df = pd.DataFrame({"ID": [1, 2, 3, 4, 50000], "a": np.random.rand(5), "b": [1.0, 2.01, 3.02, 4.03, 25.006]})
    df.to_fwf("fwf1.txt")



import numpy as np
import time


def setdiff_nd(a1, a2):
    """
    python 使用numpy求二维数组的差集
    :param a1:
    :param a2:
    :return:
    """
    # a1 = index_value
    # a2 = np.array([delta_ii_xy])
    a1_rows = a1.view([('', a1.dtype)] * a1.shape[1])
    a2_rows = a2.view([('', a2.dtype)] * a2.shape[1])

    a3 = np.setdiff1d(a1_rows, a2_rows).view(a1.dtype).reshape(-1, a1.shape[1])
    return a3


def kc_coord_3d(delta_ii_xy, xm, ym, zm, r):
    """
    :param delta_ii_xy: 当前点坐标(x,y,z)
    :param xm: size_x
    :param ym: size_y
    :param zm: size_z
    :param r: 2 * r + 1
    :return:
    返回delta_ii_xy点r邻域的点坐标
    """
    it = delta_ii_xy[0]
    jt = delta_ii_xy[1]
    kt = delta_ii_xy[2]

    xyz_min = np.array([[1, it - r], [1, jt - r], [1, kt - r]])
    xyz_min = xyz_min.max(axis=1)

    xyz_max = np.array([[xm, it + r], [ym, jt + r], [zm, kt + r]])
    xyz_max = xyz_max.min(axis=1)

    x_arange = np.arange(xyz_min[0], xyz_max[0] + 1)
    y_arange = np.arange(xyz_min[1], xyz_max[1] + 1)
    v_arange = np.arange(xyz_min[2], xyz_max[2] + 1)

    [p_k, p_i, p_j] = np.meshgrid(x_arange, y_arange, v_arange, indexing='ij')
    Index_value = np.column_stack([p_k.flatten(), p_i.flatten(), p_j.flatten()])


    Index_value = setdiff_nd(Index_value, np.array([delta_ii_xy]))

    # t0 = time.time()
    ordrho_jj = np.matmul(Index_value - 1, np.array([[1], [xm], [ym * xm]]))
    ordrho_jj.reshape([1, ordrho_jj.shape[0]])
    # t1 = time.time()
    # print((t1-t0) * 1000000000)
    # t0 = time.time()
    # ordrho_jj = np.array([((item[2] - 1) * ym * xm + (item[1] - 1) * xm + item[0] - 1) for item in Index_value])
    # t1 = time.time()
    # print((t1 - t0) * 1000000000)
    return ordrho_jj[:, 0], Index_value


def kc_coord_2d(delta_ii_xy, xm, ym, r):
    """

    :param delta_ii_xy: 当前点坐标(x,y)
    :param xm: size_x
    :param ym: size_y
    :param r: 2 * r + 1
    :return:
    返回delta_ii_xy点r邻域的点坐标
    """
    it = delta_ii_xy[0]
    jt = delta_ii_xy[1]


    p_i, p_j = np.mgrid[max(1, it - r): min(xm, it + r) + 1, max(1, jt - r): min(ym, jt + r) + 1]

    index_value = np.zeros([p_i.size, 2], np.int)
    index_value[:, 0] = p_i.flatten()
    index_value[:, 1] = p_j.ravel()

    index_value = np.array([item for item in index_value if not all(item == delta_ii_xy)])

    return index_value


if __name__ == '__main__':
    xm, ym, zm = 100, 80, 120
    r = 3
    delta_ii_xy = [43, 22, 109]
    t0 = time.time()
    index, index_value = kc_coord_3d(delta_ii_xy, xm, ym, zm, r)
    t1 = time.time()
    print((t1-t0) * 10000000)

    # size_x, size_y, size_z = 100, 80, 120
    # x_arange = np.arange(0, size_x)
    # y_arange = np.arange(0, size_y)
    # z_arange = np.arange(0, size_z)
    # [xx, yy, zz] = np.meshgrid(x_arange, y_arange, z_arange, indexing='ij')
    # xyz = np.column_stack([zz.flat, yy.flat, xx.flat])
    #
    # # size_x, size_y, size_z = data.shape
    # # p_i, p_j, p_k = np.mgrid[0: size_x, 0: size_y, 0: size_z]
    # p_i, p_j, p_k = np.mgrid[1: size_x + 1, 1: size_y + 1, 1: size_z + 1]
    # xx1 = np.zeros([p_i.size, 3], dtype=int)
    # xx1[:, 0] = p_k.flatten()
    # xx1[:, 1] = p_j.flatten()
    # xx1[:, 2] = p_i.flatten()



import numpy as np

def get_xyz(data):
    """
    :param data: 3D data
    :return: 3D data coordinates
    第1,2,3维数字依次递增
    """
    nim = data.ndim
    if nim == 3:
        size_x, size_y, size_z = data.shape
        x_arange = np.arange(1, size_x+1)
        y_arange = np.arange(1, size_y+1)
        z_arange = np.arange(1, size_z+1)
        [xx, yy, zz] = np.meshgrid(x_arange, y_arange, z_arange, indexing='ij')
        xyz = np.column_stack([zz.flatten(), yy.flatten(), xx.flatten()])

    else:
        """
            :param data: 2D data
            :return: 2D data coordinates
            第1,2维数字依次递增
            """
        size_x, size_y = data.shape
        x_arange = np.arange(1, size_x + 1)
        y_arange = np.arange(1, size_y + 1)
        [xx, yy] = np.meshgrid(x_arange, y_arange, indexing='ij')
        xyz = np.column_stack([yy.flatten(), xx.flatten()])
    return xyz


if __name__ == '__main__':
    data = np.zeros([120,100,80])
    xx = get_xyz(data)
    print(xx[[1,200,500,1000,-1],:])

    xx_2d = get_xyz(np.zeros([110, 70]))
    print(xx_2d[[1,6,-1],:])


from matplotlib import pyplot as plt
import pandas as pd


def make_plot(outcat_name, data, lable_num=True):
    outcat = pd.read_csv(outcat_name, sep='\t')
    ID = outcat['ID'].values
    Sum = outcat['Sum'].values
    if data.ndim == 2:
        fig, ax0 = plt.subplots(1, 1, figsize=(8, 6))
        ax0.imshow(data, cmap='gray')
        if outcat is not None:
            Cen1 = outcat['Cen1'] - 1
            Cen2 = outcat['Cen2'] - 1
            Peak1 = outcat['Peak1'] - 1
            Peak2 = outcat['Peak2'] - 1

            ax0.plot(Cen1, Cen2, '.', color='red')
            ax0.plot(Peak1, Peak2, '*', color='green')
            for i in range(outcat.shape[0]):

                ax0.text(Cen1[i], Cen2[i], '%d:%.2f' % (ID[i], Sum[i]), color='r', s=100)
                ax0.text(Cen1[i], Cen2[i], '%d' % (ID[i]), color='r')
                ax0.text(Peak1[i], Peak2[i], '%d' % (ID[i]), color='r')

        fig.tight_layout()
        plt.xticks([])
        plt.yticks([])
        plt.show()

    elif data.ndim == 3:
        fig, (ax0, ax1, ax2) = plt.subplots(1, 3, figsize=(8, 6))
        ax0.imshow(data.sum(axis=0), cmap='gray')
        ax1.imshow(data.sum(axis=1), cmap='gray')
        ax2.imshow(data.sum(axis=2), cmap='gray')
        if outcat is not None:
            Cen1 = outcat['Cen1'] - 1
            Cen2 = outcat['Cen2'] - 1
            Cen3 = outcat['Cen3'] - 1
            outcat = outcat - 1

            ax0.scatter(Cen1, Cen2, marker='.', c='red', s=8)
            ax1.scatter(Cen1, Cen3, marker='*', c='red', s=8)
            ax2.scatter(Cen2, Cen3, marker='^', c='red', s=8)

            if lable_num:
                for i in range(outcat.shape[0]):
                    ax0.text(Cen1[i], Cen2[i], '%d' % ID[i], color='g')
                    ax1.text(Cen1[i], Cen3[i], '%d' % ID[i], color='g')
                    ax2.text(Cen2[i], Cen3[i], '%d' % ID[i], color='g')

        fig.tight_layout()
        plt.xticks([])
        plt.yticks([])
        plt.show()


def make_tri_plot(data):
    fig, (ax0, ax1, ax2) = plt.subplots(1, 3, figsize=(8, 6))
    ax0.imshow(data.sum(axis=0), cmap='gray',)
    ax1.imshow(data.sum(axis=1), cmap='gray')
    ax2.imshow(data.sum(axis=2), cmap='gray')

    fig.tight_layout()
    plt.xticks([])
    plt.yticks([])
    plt.show()


def make_two_plot(data, outcat):
    fig, ax0 = plt.subplots(1, 1, figsize=(8, 6))
    ax0.imshow(data, cmap='gray')
    ax0.plot(outcat[:, 3] - 1, outcat[:, 4] - 1, 'r*')

    fig.tight_layout()
    plt.xticks([])
    plt.yticks([])
    plt.show()


def plot_match():
    match_txt = r'test_data\2d_simulated_clump\gaussian_out_360\match\Match_table\Match_LDC.txt'
    match = pd.read_csv(match_txt, sep='\t')
    plt.subplot(3, 2, 1)
    plt.plot(match['s_Cen1'].values, match['f_Cen1'].values, '.')
    plt.subplot(3, 2, 2)
    plt.plot(match['s_Cen2'].values, match['f_Cen2'].values, '.')
    plt.subplot(3, 2, 3)
    plt.plot(match['s_Size1'].values, match['f_Size1'].values, '.')
    plt.subplot(3, 2, 4)
    plt.plot(match['s_Size2'].values, match['f_Size2'].values, '.')
    plt.subplot(3, 2, 5)
    plt.plot(match['s_Peak'].values, match['f_Peak'].values, '.')
    plt.subplot(3, 2, 6)
    plt.plot(match['s_Sum'].values, match['f_Sum'].values, '.')
    plt.show()



from setuptools import setup


def readme_file():
      with open("README.md", encoding="utf-8") as wb:
            return wb.read()

setup(name='DensityClust',
      version='1.0.3',
      author='Luo Xiaoyu',
      description='the local density clustering algorithm',
      author_email='vastlxy@163.com',
      packages=['DensityClust']
      )